﻿using System;
using System.IO;

namespace BeeDevelopment.BbcBasicEditor {
	public class BbcBasicFile {

		#region Types

		/// <summary>
		/// Defines file attributes.
		/// </summary>
		[Flags()]
		public enum FileAttributes {
			/// <summary>No attribute flags are set.</summary>
			None = 0x00,
			/// <summary>The file is not readable by you.</summary>
			NotReadableByYou = 0x01,
			/// <summary>The file is not writable by you.</summary>
			NotWritableByYou = 0x02,
			/// <summary>The file is not executable by you.</summary>
			NotExecutableByYou = 0x04,
			/// <summary>The file is not deletable by you.</summary>
			NotDeletableByYou = 0x08,
			/// <summary>The file is not readable by others.</summary>
			NotReadableByOthers = 0x10,
			/// <summary>The file is not writable by others.</summary>
			NotWritableByOthers = 0x20,
			/// <summary>The file is not executable by others.</summary>
			NotExecutableByOthers = 0x40,
			/// <summary>The file is not deletable by others.</summary>
			NotDeletableByOthers = 0x80,
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the internal name of the file.
		/// </summary>
		public string Name { get; set; }

		/// <summary>
		/// Gets or sets the attributes of the file.
		/// </summary>
		public FileAttributes Attributes { get; set; }

		/// <summary>
		/// Gets or sets the underlying data as an array of bytes.
		/// </summary>
		public byte[] Data { get; set; }

		/// <summary>
		/// Gets or sets the load address of the file.
		/// </summary>
		public uint LoadAddress { get; set; }

		/// <summary>
		/// Gets or sets the execution address of the file.
		/// </summary>
		public uint ExecutionAddress { get; set; }

		#endregion

		#region Methods

		/// <summary>
		/// Sets the file's data from a string.
		/// </summary>
		/// <param name="s">The string to set the data with.</param>
		public void SetDataFromString(string s) {
			this.Data = Tokeniser.Tokenise(s.Split('\n'));
		}

		/// <summary>
		/// Gets the file's data as a string.
		/// </summary>
		/// <returns>The data as a string.</returns>
		public string GetDataAsString() {
			try {
				return Tokeniser.Detokenise(this.Data);
			} catch {
				throw new InvalidDataException("Invalid BBC BASIC program.");
			}
		}

		/// <summary>
		/// Saves the file to a stream.
		/// </summary>
		/// <param name="stream">The stream to save to.</param>
		private void Save(Stream stream, TexasInstruments.Types.CalculatorModel? model, Type variableType) {

			if (model.HasValue) {
				byte[] OutputData;

				using (var OutputWriter = new BinaryWriter(new MemoryStream())) {
					OutputWriter.Write(this.Data ?? new byte[0]);
					OutputData = (OutputWriter.BaseStream as MemoryStream).ToArray();
				}

				TexasInstruments.Types.Variables.Variable Var = null;

				if (variableType == typeof(TexasInstruments.Types.Variables.AppVar)) {
					Var = new TexasInstruments.Types.Variables.AppVar() {
						Name = this.Name,
						Data = OutputData,
						Model = model.Value,
					};
				} else {
					Var = new TexasInstruments.Types.Variables.ProtectedProgram() {
						Name = this.Name,
						Data = OutputData,
						Model = model.Value,
					};
				}
				
				var Group = new TexasInstruments.Types.GroupFile(model.Value);
				Group.Variables.Add(Var);
				Group.Save(stream);
			} else {
				if (this.Data != null) {
					stream.Write(this.Data, 0, this.Data.Length);
				}
			}
		}

		/// <summary>
		/// Saves the file to a file.
		/// </summary>
		/// <param name="stream">The filename to save to.</param>
		public void Save(string filename) {
			using (var s = File.Create(filename)) {
				TexasInstruments.Types.CalculatorModel? Model = null;
				var VariableType = typeof(TexasInstruments.Types.Variables.ProtectedProgram);
				switch (Path.GetExtension(filename)) {
					case ".8xp":
						Model = TexasInstruments.Types.CalculatorModel.TI83Plus;
						break;
					case ".8xv":
						Model = TexasInstruments.Types.CalculatorModel.TI83Plus;
						VariableType = typeof(TexasInstruments.Types.Variables.AppVar);
						break;
					case ".83p":
						Model = TexasInstruments.Types.CalculatorModel.TI83;
						break;
					case ".82p":
						Model = TexasInstruments.Types.CalculatorModel.TI82;
						break;
					case ".bbc":
						break;
					default:
						using (TextWriter t = new StreamWriter(s)) t.Write(this.GetDataAsString());
						return;
				}
				this.Save(s, Model, VariableType);
			}
		}

		#endregion

		#region Constructors

		public BbcBasicFile() {
			this.Data = new byte[0];
		}

		public static BbcBasicFile FromStream(Stream stream) {
			var Return = new BbcBasicFile();
			var Group = TexasInstruments.Types.GroupFile.FromStream(stream);
			foreach (var Var in Group.Variables) {
				if (Var.RawHeader.Length > 0 && (Var.RawHeader[0] == 5 || Var.RawHeader[0] == 6)) {
					var Prog = new TexasInstruments.Types.Variables.ProtectedProgram();
					Prog.RawData = Var.RawData;
					Return.Data = Prog.Data;
					return Return;
				}
			}
			throw new InvalidDataException("No usable BBC BASIC file found.");
		}

		public static BbcBasicFile FromFile(string filename) {
			using (var s = File.OpenRead(filename)) return BbcBasicFile.FromStream(s);
		}

		#endregion


	}
}
