﻿using System;
using System.Drawing;
using System.IO;
using System.Windows.Forms;
using System.Xml;
using System.Security.Cryptography;
using System.Text;
using System.Threading;
using Szotar.WindowsForms;

namespace BeeDevelopment.BbcBasicEditor {
	public partial class MainInterface : Form {

		public MainInterface() {
			InitializeComponent();
			ToolStripManager.Renderer = new ToolStripAeroRenderer(ToolbarTheme.Toolbar);
			this.Editor.ConfigurationManager.Language = "";
			this.Editor.AutoComplete.List.Clear();
			this.Editor.AutoComplete.List.AddRange(Tokeniser.GetTokenNames());
			this.CurrentFile = new BbcBasicFile();
			this.UpdateTitle();
			this.UpdateFont();
			this.WordWrap = Properties.Settings.Default.WordWrap;
			this.LoadLocation();
		}

		private void MainInterface_Load(object sender, EventArgs e) {
			this.OpenDialog.Filter = Properties.Settings.Default.DialogOpenFilter;
			this.SaveDialog.Filter = Properties.Settings.Default.DialogSaveFilter;
			//ToolStripManager.LoadSettings(this);
			this.UpdateCursorStatus();
			this.FilesizeCalculator.RunWorkerAsync("");
			if (Program.CommandLineArguments != null && Program.CommandLineArguments.Length > 0) {
				this.Open(Program.CommandLineArguments[0]);
			}
		}

		private BbcBasicFile CurrentFile { get; set; }

		#region Text Editing File Stuff

		private string currentFilename;
		/// <summary>
		/// Gets or sets the name of the file currently being edited.
		/// </summary>
		public string CurrentFilename {
			get { return this.currentFilename; }
			set {
				this.currentFilename = value;
				this.UpdateTitle();
			}
		}

		private bool currentFileIsDirty;
		/// <summary>
		/// Gets or sets whether the current file is dirty.
		/// </summary>
		public bool CurrentFileIsDirty {
			get { return this.currentFileIsDirty; }
			set {
				this.currentFileIsDirty = value;
				this.UpdateTitle();
			}
		}

		/// <summary>
		/// Updates the title of the form.
		/// </summary>
		private void UpdateTitle() {
			string Title = "";
			if (this.CurrentFilename != null) {
				Title = Path.GetFileName(this.CurrentFilename);
			} else {
				Title = "Untitled";
			}
			if (this.CurrentFileIsDirty) Title += "*";
			Title += " - " + Application.ProductName;
			this.Text = Title;
		}

		/// <summary>
		/// Creates a new file.
		/// </summary>
		public void New() {
			if (this.CurrentFileIsDirty) {
				switch (MessageBox.Show(this, "Would you like to save the current file before creating a new one?", Application.ProductName, MessageBoxButtons.YesNoCancel)) {
					case DialogResult.Cancel:
						return;
					case DialogResult.Yes:
						if (!this.Save()) return;
						break;
					case DialogResult.No:
						break;
				}
			}
			this.Editor.Text = "";
			this.CurrentFilename = null;
			this.CurrentFileIsDirty = false;
		}

		/// <summary>
		/// Saves a file.
		/// </summary>
		/// <returns>True if the file was saved, false otherwise.</returns>
		public bool Save() {
			if (this.CurrentFilename == null) {
				return this.SaveAs();
			} else {
				try {
					this.WriteAllText(this.CurrentFilename, this.Editor.Text);
					this.CurrentFileIsDirty = false;
					return true;
				} catch (Exception ex) {
					if (MessageBox.Show(this, "Error saving file. " + ex.Message + Environment.NewLine + "Would you like to try saving in a different location?", Application.ProductName, MessageBoxButtons.YesNo, MessageBoxIcon.Error) == DialogResult.Yes) {
						return this.SaveAs();
					} else {
						return false;
					}
				}
			}
		}

		/// <summary>
		/// Saves a file as...
		/// </summary>
		/// <returns>True if the file was saved, false otherwise.</returns>
		public bool SaveAs() {
			if (this.SaveDialog.ShowDialog(this) == DialogResult.Cancel) {
				return false;
			}
			try {
				this.WriteAllText(this.SaveDialog.FileName, this.Editor.Text);
				this.CurrentFilename = this.SaveDialog.FileName;
				this.CurrentFileIsDirty = false;
				return true;
			} catch (Exception ex) {
				if (MessageBox.Show(this, "Error saving file. " + ex.Message + Environment.NewLine + "Would you like to try saving in a different location?", Application.ProductName, MessageBoxButtons.YesNo, MessageBoxIcon.Error) == DialogResult.Yes) {
					return this.SaveAs();
				} else {
					return false;
				}
			}
		}

		/// <summary>
		/// Opens a file, prompting for a filename.
		/// </summary>
		public void Open() {
			this.Open(null);
		}

		/// <summary>
		/// Opens a file.
		/// </summary>
		/// <param name="filename">The name of the file to open. Passing null prompts to open using a common file dialog.</param>
		public void Open(string filename) {
			if (this.CurrentFileIsDirty) {
				switch (MessageBox.Show(this, "Would you like to save the current file before opening another?", Application.ProductName, MessageBoxButtons.YesNoCancel)) {
					case DialogResult.Cancel:
						return;
					case DialogResult.Yes:
						if (!this.Save()) return;
						break;
					case DialogResult.No:
						break;
				}
			}
			try {
				if (filename == null && this.OpenDialog.ShowDialog(this) == DialogResult.Cancel) return;
				filename = filename ?? this.OpenDialog.FileName;
				this.Editor.Text = this.ReadAllText(filename);
				this.CurrentFilename = filename;
				this.CurrentFileIsDirty = false;
				this.UpdateCursorStatus();
			} catch (Exception ex) {
				MessageBox.Show(this, "Error opening file: " + ex.Message, Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
		}

		private void NewMenu_Click(object sender, System.EventArgs e) {
			this.New();
		}

		private void OpenMenu_Click(object sender, EventArgs e) {
			this.Open();
		}

		private void SaveMenu_Click(object sender, EventArgs e) {
			this.Save();
		}

		private void SaveAsMenu_Click(object sender, EventArgs e) {
			this.SaveAs();
		}

		#endregion

		#region Text Editor Stuff

		/// <summary>
		/// Gets or sets whether word wrap is enabled.
		/// </summary>
		public bool WordWrap {
			get { return this.Editor.LineWrap.Mode != ScintillaNet.WrapMode.None; }
			set { this.Editor.LineWrap.Mode = value ? ScintillaNet.WrapMode.Word : ScintillaNet.WrapMode.None; }
		}

		/// <summary>
		/// Cuts the selected text into the clipboard.
		/// </summary>
		public void Cut() { this.Editor.Clipboard.Cut(); }
		/// <summary>
		/// Copies the selected text into the clipboard.
		/// </summary>
		public void Copy() { this.Editor.Clipboard.Copy(); }
		/// <summary>
		/// Replaces the current selection with the clipboard contents.
		/// </summary>
		public void Paste() { this.Editor.Clipboard.Paste(); }
		/// <summary>
		/// Clears the selection.
		/// </summary>
		public void Delete() { this.Editor.Selection.Clear(); }

		/// <summary>
		/// Selects all the text.
		/// </summary>
		public void SelectAll() { this.Editor.Selection.SelectAll(); }

		/// <summary>
		/// Undoes the last operation.
		/// </summary>
		public void Undo() { this.Editor.UndoRedo.Undo(); }
		/// <summary>
		/// Redoes the last operation.
		/// </summary>
		public void Redo() { this.Editor.UndoRedo.Redo(); }

		/// <summary>
		/// Shows the GoTo dialog.
		/// </summary>
		public void GoTo() { this.Editor.GoTo.ShowGoToDialog(); }

		/// <summary>
		/// Shows the quick find dialog.
		/// </summary>
		public void QuickFind() { this.Editor.FindReplace.ShowFind(); }

		/// <summary>
		/// Shows the quick replace dialog.
		/// </summary>
		public void QuickReplace() { this.Editor.FindReplace.ShowReplace(); }

		/// <summary>
		/// Updates the cursor position in the status bar.
		/// </summary>
		private void UpdateCursorStatus() {
			this.StatusCursorPosition.Text = string.Format("[{0}:{1}] : {2}", this.Editor.Lines.Current.Number + 1, this.Editor.Lines.Current.SelectionStartPosition - this.Editor.Lines.Current.StartPosition + 1, this.Editor.Lines.Count);
		}

		#endregion

		public void Renumber() {
			using (var Renumber = new RenumberForm()) {
				switch (Renumber.ShowDialog(this)) {
					case DialogResult.OK: {
							try {
								var Program = Tokeniser.Tokenise(this.Editor.Text);
								Tokeniser.Renumber(Program, Renumber.Start, Renumber.Step);
								this.Editor.Text = Tokeniser.Detokenise(Program);
								this.UpdateCursorStatus();
							} catch (Exception ex) {
								MessageBox.Show(this, "Could not renumber program: " + Environment.NewLine + ex.Message, Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Warning);
							}
						} break;
				}
			}
		}

		private void ExitMenu_Click(object sender, System.EventArgs e) {
			this.Close();
		}

		
		private void MainInterface_FormClosing(object sender, FormClosingEventArgs e) {
			if (this.CurrentFileIsDirty) {
				switch (MessageBox.Show(this, "Would you like to save the current file before exiting?", Application.ProductName, MessageBoxButtons.YesNoCancel)) {
					case DialogResult.Cancel:
						e.Cancel = true;
						return;
					case DialogResult.Yes:
						if (!this.Save()) {
							e.Cancel = true;
							return;
						}
						break;
					case DialogResult.No:
						break;
				}
			}
			this.SaveLocation();
			ToolStripManager.SaveSettings(this);
		}

		#region Fonts

		private void UpdateFont() {
			this.Editor.Font.Dispose();
			this.Editor.Font = new Font(Properties.Settings.Default.FontFamily, Properties.Settings.Default.FontSize, Properties.Settings.Default.FontStyle);
		}
		private void FontMenu_Click(object sender, EventArgs e) {
			this.FontDialog.Font = this.Editor.Font;
			if (this.FontDialog.ShowDialog(this) == DialogResult.OK) {
				Properties.Settings.Default.FontFamily = this.FontDialog.Font.FontFamily.Name;
				Properties.Settings.Default.FontSize = this.FontDialog.Font.Size;
				Properties.Settings.Default.FontStyle = this.FontDialog.Font.Style;
				Properties.Settings.Default.Save();
				this.UpdateFont();
			}
		}


		#endregion

		#region Word Wrap

		private void FormatMenu_DropDownOpening(object sender, EventArgs e) {
			this.WordWrapMenu.Checked = this.WordWrap;
		}
		private void WordWrapMenu_Click(object sender, EventArgs e) {
			Properties.Settings.Default.WordWrap = (this.WordWrap ^= true);
			Properties.Settings.Default.Save();
		}

		#endregion	

		#region Layout

		private Size CurrentSize;
		private Point CurrentLocation;

		private void MainInterface_Resize(object sender, EventArgs e) {
			if (this.WindowState == FormWindowState.Normal) CurrentSize = this.ClientSize;
		}
		private void MainInterface_LocationChanged(object sender, EventArgs e) {
			if (this.WindowState == FormWindowState.Normal) CurrentLocation = this.Location;
		}

		private void LoadLocation() {
			if (Properties.Settings.Default.LayoutPosition == new Point(-1000, -1000)) {
				this.StartPosition = FormStartPosition.WindowsDefaultBounds;
				this.CenterToScreen();
			} else {
				this.StartPosition = FormStartPosition.Manual;
				this.Location = Properties.Settings.Default.LayoutPosition;
				this.ClientSize = Properties.Settings.Default.LayoutSize;
			}
		}

		private void SaveLocation() {
			Properties.Settings.Default.LayoutSize = this.CurrentSize;
			Properties.Settings.Default.LayoutPosition = this.CurrentLocation;
			Properties.Settings.Default.Save();
		}

		#endregion

		#region File Readers/Writers
		
		public void WriteAllText(string filename, string text) {
			this.CurrentFile.SetDataFromString(text ?? "");
			this.CurrentFile.Name = Path.GetFileNameWithoutExtension(filename).ToUpperInvariant();
			this.CurrentFile.Save(filename);
		}

		public string ReadAllText(string filename) {
			try {
				this.CurrentFile = BbcBasicFile.FromFile(filename);
				return this.CurrentFile.GetDataAsString();
			} catch {
				try {
					return Tokeniser.Detokenise(File.ReadAllBytes(filename));
				} catch {
					return File.ReadAllText(filename);
				}
			}
		}

		#endregion

		#region Edit Menu

		private void UndoMenu_Click(object sender, EventArgs e) {
			this.Undo();
		}

		private void RedoMenu_Click(object sender, EventArgs e) {
			this.Redo();
		}

		private void CutMenu_Click(object sender, EventArgs e) {
			this.Cut();
		}

		private void CopyMenu_Click(object sender, EventArgs e) {
			this.Copy();
		}

		private void PasteMenu_Click(object sender, EventArgs e) {
			this.Paste();
		}

		private void DeleteMenu_Click(object sender, EventArgs e) {
			this.Delete();
		}

		private void SelectAllMenu_Click(object sender, EventArgs e) {
			this.SelectAll();
		}

		private void FindMenu_Click(object sender, EventArgs e) {
			this.QuickFind();
		}

		private void ReplaceMenu_Click(object sender, EventArgs e) {
			this.QuickReplace();
		}

		private void GoToMenu_Click(object sender, EventArgs e) {
			this.GoTo();
		}

		#endregion

		#region Tools Menu

		private void RenumberMenu_Click(object sender, EventArgs e) {
			this.Renumber();
		}

		private void ConvertToCalculatorMenu_Click(object sender, EventArgs e) {
			if (this.ConvertOpenDialog.ShowDialog(this) == DialogResult.OK && this.ConvertOpenDialog.FileNames.Length > 0) {

				this.SaveDialog.FileName = Path.Combine(Path.GetDirectoryName(this.ConvertOpenDialog.FileNames[0]), Path.ChangeExtension(this.ConvertOpenDialog.FileNames[0], ".8xp"));
				this.ConvertSaveFolderDialog.SelectedPath = Path.GetDirectoryName(this.ConvertOpenDialog.FileNames[0]);

				if ((this.ConvertOpenDialog.FileNames.Length > 1 ? this.ConvertSaveFolderDialog.ShowDialog(this) : this.SaveDialog.ShowDialog(this)) == DialogResult.OK) {
					try {
						foreach (var ToConvert in this.ConvertOpenDialog.FileNames) {
							var Converted = new BbcBasicFile();
							Converted.Name = Path.GetFileNameWithoutExtension(ToConvert).ToUpperInvariant();
							Converted.Data = File.ReadAllBytes(ToConvert);
							if (this.ConvertOpenDialog.FileNames.Length == 1) {
								Converted.Save(this.SaveDialog.FileName);
							} else {
								var DestName = Path.Combine(Path.GetDirectoryName(ToConvert), Path.ChangeExtension(ToConvert, ".8xp"));
								if (!File.Exists(DestName) || MessageBox.Show(this, "The file " + DestName + " already exists. Would you like to overwrite it?", "Convert", MessageBoxButtons.YesNo, MessageBoxIcon.None, MessageBoxDefaultButton.Button2) == DialogResult.Yes) {
									Converted.Save(DestName);
								}
							}
						}
					} catch (Exception ex) {
						MessageBox.Show(this, "Error converting file: " + ex.Message, Application.ProductName, MessageBoxButtons.OK);
					}
				}
			}
		}

		#endregion


		private void ExportHtmlMenuItem_Click(object sender, EventArgs e) {
			if (this.ExportHtmlSaveDialog.ShowDialog(this) == DialogResult.OK) {
				XmlWriter HtmlOut = XmlWriter.Create(this.ExportHtmlSaveDialog.FileName, new XmlWriterSettings() {
					Indent = true,
					OmitXmlDeclaration = true,
				});
				try {
					HtmlOut.WriteDocType("html", "-//W3C//DTD XHTML 1.0 Strict//EN", "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd", null);

					HtmlOut.WriteStartElement("html", "http://www.w3.org/1999/xhtml");

					HtmlOut.WriteStartElement("head");
					HtmlOut.WriteElementString("title", string.IsNullOrEmpty(this.CurrentFilename) ? "BBC BASIC Program" : Path.GetFileNameWithoutExtension(this.CurrentFilename));
					HtmlOut.WriteStartElement("style");
					HtmlOut.WriteAttributeString("type", "text/css");
					HtmlOut.WriteRaw(Properties.Settings.Default.ExportHtmlStyle);
					HtmlOut.WriteEndElement(); // style
					HtmlOut.WriteEndElement(); // head

					HtmlOut.WriteStartElement("body");
					HtmlOut.WriteStartElement("div");
					HtmlOut.WriteAttributeString("class", "code");

					string s;
					HtmlOut.WriteRaw(s = Tokeniser.Detokenise(Tokeniser.Tokenise(this.Editor.Text), Tokeniser.DetokeniserSettings.Html));

					HtmlOut.WriteEndElement(); // div
					HtmlOut.WriteEndElement(); // body
					HtmlOut.WriteEndElement(); // html
				} catch (Exception ex) {
					MessageBox.Show(this, "Error exporting HTML: " + ex.Message, Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
				} finally {
					HtmlOut.Close();
				}
			}
			
		}

		private void BeautifyCodeMenuItem_Click(object sender, EventArgs e) {
			try {
				this.Editor.Text = Tokeniser.Detokenise(Tokeniser.Tokenise(this.Editor.Text));
				this.UpdateCursorStatus();
			} catch (Exception ex) {
				MessageBox.Show(this,"Could not beautify code: " + ex.Message, Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
			}
		}

		private void HelpContents_Click(object sender, EventArgs e) {
			this.ShowHelp(HtmlHelpViewer.Command.HH_DISPLAY_TOC, null);
		}

		private void HelpIndex_Click(object sender, EventArgs e) {
			this.ShowHelp(HtmlHelpViewer.Command.HH_DISPLAY_INDEX, null);
		}

		private void AboutBbcBasicMenu_Click(object sender, EventArgs e) {
			this.ShowHelp(HtmlHelpViewer.Command.HH_DISPLAY_TOPIC, "::/Credits.htm");
		}

		private void ShowHelp(HtmlHelpViewer.Command command, object args) {
			if (HtmlHelpViewer.HtmlHelp(this.Handle, Path.Combine(Application.StartupPath, Properties.Settings.Default.HelpFile) + args, command, null) == IntPtr.Zero) {
				if (MessageBox.Show(this, "Help file '" + Path.GetFileName(Properties.Settings.Default.HelpFile) + "' could not be found." + Environment.NewLine + Environment.NewLine + "Would you like to browse for it?", Application.ProductName, MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation) == DialogResult.Yes) {
					if (this.OpenChmDialog.ShowDialog(this) == DialogResult.OK) {
						Properties.Settings.Default.HelpFile = this.OpenChmDialog.FileName;
						Properties.Settings.Default.Save();
						this.ShowHelp(command, args);
					}
				}
			}
		}


		#region Toolbar Buttons

		private void NewButton_Click(object sender, EventArgs e) {
			this.New();
		}

		private void OpenButton_Click(object sender, EventArgs e) {
			this.Open();
		}

		private void SaveButton_Click(object sender, EventArgs e) {
			this.Save();
		}

		private void CutButton_Click(object sender, EventArgs e) {
			this.Cut();
		}

		private void CopyButton_Click(object sender, EventArgs e) {
			this.Copy();
		}

		private void PasteButton_Click(object sender, EventArgs e) {
			this.Paste();
		}

		private void UndoButton_Click(object sender, EventArgs e) {
			this.Undo();
		}

		private void RedoButton_Click(object sender, EventArgs e) {
			this.Redo();
		}

		private void FindButton_Click(object sender, EventArgs e) {
			this.QuickFind();
		}

		private void ReplaceButton_Click(object sender, EventArgs e) {
			this.QuickReplace();
		}

		private void RenumberButton_Click(object sender, EventArgs e) {
			this.Renumber();
		}

		#endregion

		private void Editor_TextInserted(object sender, ScintillaNet.TextModifiedEventArgs e) {
			this.CurrentFileIsDirty = true;
		}
		private void Editor_TextDeleted(object sender, ScintillaNet.TextModifiedEventArgs e) {
			this.CurrentFileIsDirty = true;
		}

		private void MainStripContainer_Layout(object sender, LayoutEventArgs e) {
			this.MainStripContainer.ContentPanel.Padding = new Padding(
				this.MainStripContainer.LeftToolStripPanel.Controls.Count > 0 ? 1 : 0,
				this.MainStripContainer.TopToolStripPanel.Controls.Count > 0 ? 1 : 0,
				this.MainStripContainer.RightToolStripPanel.Controls.Count > 0 ? 1 : 0,
				this.MainStripContainer.BottomToolStripPanel.Controls.Count > 0 ? 1 : 0
			);
		}

		private void Editor_FileDrop(object sender, ScintillaNet.FileDropEventArgs e) {
			if (e.FileNames.Length == 1) {
				this.Open(e.FileNames[0]);
			}
		}

		private void Editor_SelectionChanged(object sender, EventArgs e) {
			this.UpdateCursorStatus();
		}

		private void FilesizeCalculator_RunWorkerCompleted(object sender, System.ComponentModel.RunWorkerCompletedEventArgs e) {
			if (!e.Cancelled) {
				var SizeLabel = "";
				if (e.Result != null) {
					decimal Size = (int)e.Result;
					var Suffixes = new[] { " bytes", "KB", "MB", "GB" };
					int SuffixIndex = 0;
					while (Size >= 1024m && SuffixIndex < Suffixes.Length - 1) {
						Size /= 1024m;
						++SuffixIndex;
					}
					SizeLabel = string.Format("{0:##.##}{1}", Size, Suffixes[SuffixIndex]);
				}
				this.Invoke(new MethodInvoker(() => { this.StatusBinarySize.Text = SizeLabel; }));
			}
		}

		private void FilesizeCalculator_DoWork(object sender, System.ComponentModel.DoWorkEventArgs e) {
			e.Result = null;
			int Filesize = 0;
			foreach (var line in ((string)e.Argument).Split('\n')) {
				if (this.FilesizeCalculator.CancellationPending) {
					e.Cancel = true;
					return;
				}
				try {
					Filesize += Tokeniser.Tokenise(line).Length;
				} catch {
					return;
				}
			}
			e.Result = Filesize;
		}

		private void Editor_DocumentChange(object sender, ScintillaNet.NativeScintillaEventArgs e) {
			if ((e.SCNotification.modificationType & (0x01 | 0x02 | 0x100)) == 0) return;
			lock (this.FilesizeCalculator) {
				if (this.FilesizeCalculator.IsBusy) {
					this.FilesizeCalculator.CancelAsync();
					while (this.FilesizeCalculator.IsBusy) {
						Thread.Sleep(0);
						Application.DoEvents();
					}
				}
				this.FilesizeCalculator.RunWorkerAsync(this.Editor.Text);
			}
		}

	}
}
