﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Text;
using System.IO;

namespace BeeDevelopment.BbcBasicEditor {

	/// <summary>
	/// Provides methods for tokenising and detokenising BBC BASIC files.
	/// </summary>
	public static class Tokeniser {


		static bool IsValidInHexConstant(this char c) {
			return (c >= 'A' && c <= 'Z') || (c >= '0' && c <= '9');
		}

		static bool IsValidInName(this char c) {
			return (c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || (c >= '0' && c <= '9') || c == '_' || c == '@';
		}

		static bool IsValidInLineNumberList(this char c) {
			return (c >= '0' && c <= '9') || char.IsWhiteSpace(c) || c == ',';
		}

		static byte[] ToEncodedLineNumber(this ushort line) {
			return new byte[]{
				0x8D,
				(byte)((((line & 0x00C0) >> 2) | ((line & 0xC000) >> 12)) ^ 0x54),
				(byte)(((line >> 0) & 0x3F) | 0x40),
				(byte)(((line >> 8) & 0x3F) | 0x40),
			};
		}

		static ushort DecodeLineNumber(this byte[] program, int offset) {
			if (program[offset] != 0x8D) throw new InvalidDataException("Not a valid line number.");
			return (ushort)((((program[offset + 1] ^ 0x54) & 0x30) << 2) |
									(((program[offset + 1] ^ 0x54) & 0x0C) << 12) |
									(program[offset + 2] & 0x3F) |
									((program[offset + 3] & 0x3F) << 8));
		}

		static bool HasStringAtOffset(this byte[] program, int offset, string str) {
			var EncodedStr = new BbcBasicEncoding().GetBytes(str);
			if (offset + EncodedStr.Length >= program.Length) return false;
			for (int i = 0; i < EncodedStr.Length; i++) {
				if (program[i + offset] != EncodedStr[i]) return false;
			}
			return true;
		}

		#region Types

		/// <summary>
		/// Represents flags that control the way a keyword is tokenised.
		/// </summary>
		[Flags()]
		private enum KeywordFlags : byte {
			/// <summary>
			/// Represents no flags.
			/// </summary>
			None = 0,
			/// <summary>
			/// If set, the keyword is not tokenised if it followed by an alphanumeric character.
			/// </summary>
			Conditional = 1 << 0,
			/// <summary>
			/// If set, tells the tokeniser to go into 'middle of statement' mode after this token.
			/// </summary>
			Middle = 1 << 1,
			/// <summary>
			/// If set, tells the tokeniser to go into 'start of statement' mode after this token.
			/// </summary>
			Start = 1 << 2,
			/// <summary>
			/// If set, tells the tokeniser not to tokenise the name following the token.
			/// </summary>
			FnProc = 1 << 3,
			/// <summary>
			/// If set, the tokeniser should start tokenising line numbers after the token.
			/// </summary>
			LineNumber = 1 << 4,
			/// <summary>
			/// If set, do not tokenise the rest of the line.
			/// </summary>
			Remark = 1 << 5,
			/// <summary>
			/// If set, add 0x40 to the token if it's at the start of the statement.
			/// </summary>
			PseudoVariable = 1 << 6,
		}

		/// <summary>
		/// Defines the current state of the tokeniser.
		/// </summary>
		enum TokeniserState {
			/// <summary>
			/// The tokeniser is at the start of the statement.
			/// </summary>
			StartOfStatement,
			/// <summary>
			/// The tokeniser is in the middle of the statement.
			/// </summary>
			MiddleOfStatement,
		}

		private struct Keyword {

			private byte token;
			/// <summary>
			/// Gets or sets the token representation of the keyword.
			/// </summary>
			public byte Token { get { return this.token; } set { this.token = value; } }

			private string text;
			/// <summary>
			/// Gets or sets the text of the token.
			/// </summary>
			public string Text { get { return this.text; } set { this.text = value; } }


			private int indent;
			/// <summary>
			/// Gets or sets the direction and quantity of indentation that this keyword modifies.
			/// </summary>
			public int Indent { get { return this.indent; } set { this.indent = value; } }

			private bool isSpecial;
			/// <summary>
			/// Gets or sets whether the keyword is "special" (only affects the colour it is highlighted).
			/// </summary>
			public bool IsSpecial { get { return this.isSpecial; } set { this.isSpecial = value; } }

			private KeywordFlags flags;
			/// <summary>
			/// Gets or sets the keyword's flags.
			/// </summary>
			public KeywordFlags Flags { get { return this.flags; } set { this.flags = value; } }

			public override string ToString() {
				return this.Text;
			}

		}

		/// <summary>
		/// Defines the settings used to format detokenised source code.
		/// </summary>
		public class DetokeniserSettings {

			/// <summary>
			/// Gets or sets the string value used for indentation.
			/// </summary>
			public string IndentString { get; set; }

			/// <summary>
			/// Gets or sets whether line numbers should be indented.
			/// </summary>
			public bool IndentLineNumbers { get; set; }

			/// <summary>
			/// Gets or sets whether the detokenised source should particular characters replaced by their HTML entities.
			/// </summary>
			public bool HtmlEntities { get; set; }

			/// <summary>
			/// Gets or sets the string to output before a keyword.
			/// </summary>
			public string KeywordPrefix { get; set; }

			/// <summary>
			/// Gets or sets the string to output after a keyword.
			/// </summary>
			public string KeywordSuffix { get; set; }

			/// <summary>
			/// Gets or sets the string to output before a "special" keyword.
			/// </summary>
			public string SpecialKeywordPrefix { get; set; }

			/// <summary>
			/// Gets or sets the string to output after a "special" keyword.
			/// </summary>
			public string SpecialKeywordSuffix { get; set; }

			/// <summary>
			/// Gets or sets the string to output before a line number.
			/// </summary>
			public string LineNumberPrefix { get; set; }
			/// <summary>
			/// Gets or sets the string to output after a line number.
			/// </summary>
			public string LineNumberSuffix { get; set; }

			/// <summary>
			/// Gets or sets the string to output before a complete line.
			/// </summary>
			public string LinePrefix { get; set; }
			/// <summary>
			/// Gets or sets the string to output after a complete line.
			/// </summary>
			public string LineSuffix { get; set; }

			/// <summary>
			/// Gets or sets the string to output before a string literal.
			/// </summary>
			public string StringPrefix { get; set; }
			/// <summary>
			/// Gets or sets the string to output after a string literal.
			/// </summary>
			public string StringSuffix { get; set; }

			/// <summary>
			/// Gets or sets the string to output before a star command.
			/// </summary>
			public string StarCommandPrefix { get; set; }
			/// <summary>
			/// Gets or sets the string to output after a star commant.
			/// </summary>
			public string StarCommandSuffix { get; set; }

			/// <summary>
			/// Gets or sets the string to output before a comment.
			/// </summary>
			public string CommentPrefix { get; set; }
			/// <summary>
			/// Gets or sets the string to output after a comment.
			/// </summary>
			public string CommentSuffix { get; set; }

			/// <summary>
			/// Creates an instance of the <see cref="DetokeniserSettings"/> class with default settings.
			/// </summary>
			public DetokeniserSettings() {
				this.IndentString = "  ";
				this.IndentLineNumbers = true;
				this.HtmlEntities = false;
				this.LineNumberSuffix = " ";
				this.LineSuffix = Environment.NewLine;
			}

			/// <summary>
			/// Returns a <see cref="DetokeniserSettings"/> instance suitable for generating an HTML file.
			/// </summary>
			public static DetokeniserSettings Html {
				get {
					return new DetokeniserSettings {
						HtmlEntities = true,
						IndentString = "&nbsp;&nbsp;",
						KeywordPrefix = @"<span class=""keyword"">",
						KeywordSuffix = @"</span>",
						SpecialKeywordPrefix = @"<span class=""keyword special"">",
						SpecialKeywordSuffix = @"</span>",
						LineNumberPrefix = @"<span class=""line-number"">",
						LineNumberSuffix = @"</span>&nbsp;",
						StringPrefix = @"<span class=""string"">",
						StringSuffix = @"</span>",
						LinePrefix = @"<div class=""line"">",
						LineSuffix = @"</div>" + Environment.NewLine,
						StarCommandPrefix = @"<span class=""star-command"">",
						StarCommandSuffix = @"</span>",
						CommentPrefix = @"<span class=""comment"">",
						CommentSuffix = @"</span>",
					};
				}
			}
		}

		#endregion

		private static Dictionary<string, Keyword> StringToKeywordMap;
		private static Dictionary<byte, Keyword> TokenToKeywordMap;
		private static List<string> KeywordStringsLongestFirst;
		private static List<Keyword> ExtendedKeywords;

		static Tokeniser() {

			var Tokens = new[]{
				new { Token = 0x80, Keyword = "AND", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x81, Keyword = "DIV", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x82, Keyword = "EOR", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x83, Keyword = "MOD", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x84, Keyword = "OR", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x85, Keyword = "ERROR", Flags = KeywordFlags.Start, Indent = 0, IsSpecial = false },
				new { Token = 0x86, Keyword = "LINE", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x87, Keyword = "OFF", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x88, Keyword = "STEP", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x89, Keyword = "SPC", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x8A, Keyword = "TAB(", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x8B, Keyword = "ELSE", Flags = KeywordFlags.LineNumber | KeywordFlags.Start, Indent = 0, IsSpecial = false },
				new { Token = 0x8C, Keyword = "THEN", Flags = KeywordFlags.LineNumber | KeywordFlags.Start, Indent = 0, IsSpecial = false },
				new { Token = 0x8D, Keyword = "$LINE$", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x8E, Keyword = "OPENIN", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x8F, Keyword = "PTR", Flags = KeywordFlags.PseudoVariable|KeywordFlags.Middle|KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0x90, Keyword = "PAGE", Flags = KeywordFlags.PseudoVariable|KeywordFlags.Middle|KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0x91, Keyword = "TIME", Flags = KeywordFlags.PseudoVariable|KeywordFlags.Middle|KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0x92, Keyword = "LOMEM", Flags = KeywordFlags.PseudoVariable|KeywordFlags.Middle|KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0x93, Keyword = "HIMEM", Flags = KeywordFlags.PseudoVariable|KeywordFlags.Middle|KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0x94, Keyword = "ABS", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x95, Keyword = "ACS", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x96, Keyword = "ADVAL", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x97, Keyword = "ASC", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x98, Keyword = "ASN", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x99, Keyword = "ATN", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x9A, Keyword = "BGET", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0x9B, Keyword = "COS", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x9C, Keyword = "COUNT", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0x9D, Keyword = "DEG", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0x9E, Keyword = "ERL", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0x9F, Keyword = "ERR", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xA0, Keyword = "EVAL", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xA1, Keyword = "EXP", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xA2, Keyword = "EXT", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xA3, Keyword = "FALSE", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xA4, Keyword = "FN", Flags = KeywordFlags.FnProc, Indent = 0, IsSpecial = false },
				new { Token = 0xA5, Keyword = "GET", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xA6, Keyword = "INKEY", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xA7, Keyword = "INSTR", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xA8, Keyword = "INT", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xA9, Keyword = "LEN", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xAA, Keyword = "LN", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xAB, Keyword = "LOG", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xAC, Keyword = "NOT", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xAD, Keyword = "OPENUP", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xAE, Keyword = "OPENOUT", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xAF, Keyword = "PI", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xB0, Keyword = "POINT(", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xB1, Keyword = "POS", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xB2, Keyword = "RAD", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xB3, Keyword = "RND", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xB4, Keyword = "SGN", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xB5, Keyword = "SIN", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xB6, Keyword = "SQR", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xB7, Keyword = "TAN", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xB8, Keyword = "TO", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xB9, Keyword = "TRUE", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xBA, Keyword = "USR", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xBB, Keyword = "VAL", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xBC, Keyword = "VPOS", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xBD, Keyword = "CHR$", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xBE, Keyword = "GET$", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xBF, Keyword = "INKEY$", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xC0, Keyword = "LEFT$(", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xC1, Keyword = "MID$(", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xC2, Keyword = "RIGHT$(", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xC3, Keyword = "STR$", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xC4, Keyword = "STRING$(", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xC5, Keyword = "EOF", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xC6, Keyword = "AUTO", Flags = KeywordFlags.LineNumber, Indent = 0, IsSpecial = false },
				new { Token = 0xC7, Keyword = "DELETE", Flags = KeywordFlags.LineNumber, Indent = 0, IsSpecial = false },
				new { Token = 0xC8, Keyword = "LOAD", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xC9, Keyword = "LIST", Flags = KeywordFlags.LineNumber, Indent = 0, IsSpecial = false },
				new { Token = 0xCA, Keyword = "NEW", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xCB, Keyword = "OLD", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xCC, Keyword = "RENUMBER", Flags = KeywordFlags.LineNumber, Indent = 0, IsSpecial = false },
				new { Token = 0xCD, Keyword = "SAVE", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xCE, Keyword = "PUT", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xCF, Keyword = "PTR", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xD0, Keyword = "PAGE", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xD1, Keyword = "TIME", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xD2, Keyword = "LOMEM", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xD3, Keyword = "HIMEM", Flags = KeywordFlags.None, Indent = 0, IsSpecial = false },
				new { Token = 0xD4, Keyword = "SOUND", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xD5, Keyword = "BPUT", Flags = KeywordFlags.Middle|KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xD6, Keyword = "CALL", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xD7, Keyword = "CHAIN", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xD8, Keyword = "CLEAR", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xD9, Keyword = "CLOSE", Flags = KeywordFlags.Middle|KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xDA, Keyword = "CLG", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xDB, Keyword = "CLS", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xDC, Keyword = "DATA", Flags = KeywordFlags.Remark, Indent = 0, IsSpecial = false },
				new { Token = 0xDD, Keyword = "DEF", Flags = KeywordFlags.None, Indent = 0, IsSpecial = true },
				new { Token = 0xDE, Keyword = "DIM", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xDF, Keyword = "DRAW", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xE0, Keyword = "END", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xE1, Keyword = "ENDPROC", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xE2, Keyword = "ENVELOPE", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xE3, Keyword = "FOR", Flags = KeywordFlags.Middle, Indent = 1, IsSpecial = false },
				new { Token = 0xE4, Keyword = "GOSUB", Flags = KeywordFlags.LineNumber|KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xE5, Keyword = "GOTO", Flags = KeywordFlags.LineNumber|KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xE6, Keyword = "GCOL", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xE7, Keyword = "IF", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xE8, Keyword = "INPUT", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xE9, Keyword = "LET", Flags = KeywordFlags.Start, Indent = 0, IsSpecial = false },
				new { Token = 0xEA, Keyword = "LOCAL", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xEB, Keyword = "MODE", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xEC, Keyword = "MOVE", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xED, Keyword = "NEXT", Flags = KeywordFlags.Middle, Indent = -1, IsSpecial = false },
				new { Token = 0xEE, Keyword = "ON", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xEF, Keyword = "VDU", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xF0, Keyword = "PLOT", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xF1, Keyword = "PRINT", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xF2, Keyword = "PROC", Flags = KeywordFlags.FnProc|KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xF3, Keyword = "READ", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xF4, Keyword = "REM", Flags = KeywordFlags.Remark, Indent = 0, IsSpecial = false },
				new { Token = 0xF5, Keyword = "REPEAT", Flags = KeywordFlags.None, Indent = 1, IsSpecial = false },
				new { Token = 0xF6, Keyword = "REPORT", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xF7, Keyword = "RESTORE", Flags = KeywordFlags.LineNumber|KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xF8, Keyword = "RETURN", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xF9, Keyword = "RUN", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xFA, Keyword = "STOP", Flags = KeywordFlags.Conditional, Indent = 0, IsSpecial = false },
				new { Token = 0xFB, Keyword = "COLOUR", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xFC, Keyword = "TRACE", Flags = KeywordFlags.LineNumber|KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xFD, Keyword = "UNTIL", Flags = KeywordFlags.Middle, Indent = -1, IsSpecial = false },
				new { Token = 0xFE, Keyword = "WIDTH", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
				new { Token = 0xFF, Keyword = "OSCLI", Flags = KeywordFlags.Middle, Indent = 0, IsSpecial = false },
			};

			Tokeniser.TokenToKeywordMap = new Dictionary<byte, Keyword>();
			Tokeniser.StringToKeywordMap = new Dictionary<string, Keyword>();
			Tokeniser.KeywordStringsLongestFirst = new List<string>();


			foreach (var Token in Tokens) {

				var Keyword = new Keyword() {
					Token = (byte)Token.Token,
					Flags = Token.Flags,
					Indent = Token.Indent,
					Text = Token.Keyword,
					IsSpecial = Token.IsSpecial,
				};

				TokenToKeywordMap.Add((byte)Token.Token, Keyword);

				if (Keyword.Text != "$LINE$") {
					if (!KeywordStringsLongestFirst.Contains(Keyword.Text)) KeywordStringsLongestFirst.Add(Keyword.Text);
					if (!StringToKeywordMap.ContainsKey(Token.Keyword)) {
						StringToKeywordMap.Add(Token.Keyword, Keyword);
					}
				}
			}

			KeywordStringsLongestFirst.Sort((a, b) => b.Length - a.Length);

			Tokeniser.ExtendedKeywords = new List<Keyword>(new [] {
				new Keyword() { Indent = 0, Text = "CIRCLE" },
				new Keyword() { Indent = 0, Text = "RECTANGLE" },
				new Keyword() { Indent = 0, Text = "ELLIPSE" },
				new Keyword() { Indent = 0, Text = "FILL" },
				new Keyword() { Indent = +1, Text = "WHILE" },
				new Keyword() { Indent = -1, Text = "ENDWHILE" },
				new Keyword() { Indent = 0, Text = "EXITREPEAT" },
				new Keyword() { Indent = 0, Text = "EXITWHILE" },
				new Keyword() { Indent = 0, Text = "EXITFOR" },
				new Keyword() { Indent = -1, Text = "EXIT" },
				new Keyword() { Indent = 0, Text = "BY" },
			});

		}

		public static string Detokenise(byte[] data) {
			return Tokeniser.Detokenise(data, new DetokeniserSettings());
		}

		private static string HtmlEntities(string s) {
			s = s.Replace("&", "&amp;");
			s = s.Replace("<", "&lt;");
			s = s.Replace(">", "&gt;");
			return s;
		}

		public static IEnumerable<string> GetTokenNames() {
			var Result = new List<string>(KeywordStringsLongestFirst);
			Result.Sort();
			return Result;
		}

		/*internal static string GenerateTokenTable() {
			var Result = new StringBuilder(1024);
			int Cols = 4;
			for (int j = 0; j < Cols; ++j) {
				Result.AppendLine(@"{|class=""wikitable"" style=""float: left;""");
				Result.AppendLine("|-");
				Result.AppendLine("! Token || Keyword || Flags");
				for (int i = (0x80 / Cols) * j; i < (0x80 / Cols) * (j + 1); i++) {
					Keyword Keyword;
					if (!TokenToKeywordMap.TryGetValue((byte)(i + 0x80), out Keyword)) Keyword = new Keyword() { Token = (byte)(i + 0x80) };
					Result.AppendLine("|-");
					string Flags = "-";
					Flags += (Keyword.Flags & KeywordFlags.PseudoVariable) == KeywordFlags.None ? "-" : "P";
					Flags += (Keyword.Flags & KeywordFlags.Remark) == KeywordFlags.None ? "-" : "R";
					Flags += (Keyword.Flags & KeywordFlags.LineNumber) == KeywordFlags.None ? "-" : "L";
					Flags += (Keyword.Flags & KeywordFlags.FnProc) == KeywordFlags.None ? "-" : "F";
					Flags += (Keyword.Flags & KeywordFlags.Start) == KeywordFlags.None ? "-" : "S";
					Flags += (Keyword.Flags & KeywordFlags.Middle) == KeywordFlags.None ? "-" : "M";
					Flags += (Keyword.Flags & KeywordFlags.Conditional) == KeywordFlags.None ? "-" : "C";
					Result.AppendLine(string.Format("| {0:X2} || {1} || <code>{2}</code>", Keyword.Token, Keyword.Text != null ? (Keyword.Text == "$LINE$" ? "''line no.''" : "<code>" + Keyword.Text + "</code>") : "''N/A''", Flags));
				}
				Result.AppendLine("|}");
			}			
			return Result.ToString();
		}*/

		public static string Detokenise(byte[] data, DetokeniserSettings settings) {
			var Result = new StringBuilder(data.Length * 4);
			var WorkingLine = new StringBuilder(256);

			var TargetEncoding = new BbcBasicEncoding();

			var Indent = 0;

			for (int i = 0; i < data.Length; ++i) {

				var LineLength = data[i++];
				var LineNumber = data[i++] + data[i++] * 256;

				if (LineLength == 0 && LineNumber == 0xFFFF) break; // End-of-file.

				Result.Append(settings.LinePrefix);

				var LineNumberString = "";
				if (LineNumber != 0) {
					LineNumberString = LineNumber.ToString(CultureInfo.InvariantCulture);
				}
				if (settings.IndentLineNumbers) {
					var LineNumberIndent = "".PadRight(5 - LineNumberString.Length, ' ');
					if (settings.HtmlEntities) LineNumberIndent = LineNumberIndent.Replace(" ", "&nbsp;");
					Result.Append(LineNumberIndent);
				}
				Result.Append(settings.LineNumberPrefix);
				Result.Append(LineNumberString);
				Result.Append(settings.LineNumberSuffix);

				WorkingLine.Length = 0;
				var OldIndent = Indent;
				for (int j = 0; j < LineLength - 4; ++j, ++i) {
					var Token = data[i];
					if ((char)Token == '"') {
						WorkingLine.Append(settings.StringPrefix);
						var InString = false;
						int iO = i, jO = j;
						int CharsFound = -1;
						while (j < LineLength - 4) {
							if (data[i] == (byte)'"') {
								InString ^= true;
							}
							WorkingLine.Append(TargetEncoding.GetChars(data, i, 1));
							++CharsFound;
							if (!InString) {
								break;
							}
							++i; ++j;
						}
						i = iO + CharsFound; j = jO + CharsFound;
						WorkingLine.Append(settings.StringSuffix);
					} else if (j == 0 && Token == (byte)'*') {
						WorkingLine.Append(settings.StarCommandPrefix);
						for (; j < LineLength - 4; ++j, ++i) {
							WorkingLine.Append(TargetEncoding.GetChars(data, i, 1));
						}
						--i; --j;
						WorkingLine.Append(settings.StarCommandSuffix);
					} else if (Token == 0xF4) {
						WorkingLine.Append(settings.CommentPrefix);
						WorkingLine.Append("REM");
						++i; ++j;
						for (; j < LineLength - 4; ++j, ++i) {
							WorkingLine.Append(TargetEncoding.GetChars(data, i, 1));
						}
						--i; --j;
						WorkingLine.Append(settings.CommentSuffix);
					} else {
						Keyword Keyword;
						if (TokenToKeywordMap.TryGetValue(Token, out Keyword)) {
							if (Keyword.Text == "$LINE$") {
								var EncodedLineNumber =
									(((data[i + 1] ^ 0x54) & 0x30) << 2) |
									(((data[i + 1] ^ 0x54) & 0x0C) << 12) |
									(data[i + 2] & 0x3F) |
									((data[i + 3] & 0x3F) << 8);
								WorkingLine.Append(EncodedLineNumber.ToString(CultureInfo.InvariantCulture));
								i += 3; j += 3;
							} else {
								WorkingLine.Append(Keyword.IsSpecial ? settings.SpecialKeywordPrefix : settings.KeywordPrefix);
								var KeywordName = settings.HtmlEntities ? HtmlEntities(Keyword.Text) : Keyword.Text;
								if (KeywordName.EndsWith("(")) {
									WorkingLine.Append(KeywordName.Substring(0, KeywordName.Length - 1));
								} else {
									WorkingLine.Append(KeywordName);
								}
								WorkingLine.Append(Keyword.IsSpecial ? settings.SpecialKeywordSuffix : settings.KeywordSuffix);
								if (KeywordName.EndsWith("(")) {
									WorkingLine.Append("(");
								}
								Indent += Keyword.Indent;
							}
						} else {

							bool FoundExtendedKeyword = false;
							foreach (var ExtendedKeyword in Tokeniser.ExtendedKeywords) {
								if (data.HasStringAtOffset(i, ExtendedKeyword.Text)) {
									WorkingLine.Append(Keyword.IsSpecial ? settings.SpecialKeywordPrefix : settings.KeywordPrefix);
									WorkingLine.Append(ExtendedKeyword.Text);
									WorkingLine.Append(Keyword.IsSpecial ? settings.SpecialKeywordSuffix : settings.KeywordSuffix);
									int ExtendedKeywordLength = TargetEncoding.GetByteCount(ExtendedKeyword.Text) - 1;
									i += ExtendedKeywordLength; j += ExtendedKeywordLength;
									Indent += ExtendedKeyword.Indent;
									FoundExtendedKeyword = true;
									break;
								}
							}

							if (!FoundExtendedKeyword) {
								string C = TargetEncoding.GetString(data, i, 1);
								WorkingLine.Append(settings.HtmlEntities ? HtmlEntities(C) : C);
							}
						}
					}
				}

				var IndentAmount = (Indent > OldIndent ? OldIndent : Indent);
				for (int indent = 0; indent < IndentAmount; ++indent) {
					Result.Append(settings.IndentString);
				}

				Result.Append(WorkingLine);

				Result.Append(settings.LineSuffix);

			}

			return Result.ToString().TrimEnd();
		}

		/// <summary>
		/// Tokenise BBC BASIC source code.
		/// </summary>
		/// <param name="s">The string to tokenise.</param>
		/// <returns>The tokenised BBC BASIC program.</returns>
		public static byte[] Tokenise(string s) {
			return Tokenise(s.Split('\n'));
		}

		/// <summary>
		/// Tokenise BBC BASIC source code.
		/// </summary>
		/// <param name="lines">The lines of BBC BASIC to tokenise.</param>
		/// <returns>The tokenised BBC BASIC program.</returns>
		public static byte[] Tokenise(string[] lines) {
			var Output = new List<byte>(lines.Length * 128);
			var LineNumberString = new StringBuilder(6);
			var TargetEncoding = new BbcBasicEncoding();

			// Iterate over all lines in the source file.
			foreach (var Line in lines) {

				// Check if the line has something in it first...
				if (string.IsNullOrEmpty(Line)) continue;
				var TrimmedLine = Line.Trim();
				if (string.IsNullOrEmpty(TrimmedLine)) continue;

				// Try and grab a line number.
				LineNumberString.Length = 0;
				int i = 0;
				for (; i < TrimmedLine.Length && char.IsDigit(TrimmedLine[i]); ++i) {
					LineNumberString.Append(TrimmedLine[i]);
				}
				var LineNumber = LineNumberString.Length == 0 ? 0 : ushort.Parse(LineNumberString.ToString(), CultureInfo.InvariantCulture);

				// Remove the start of the string if need be.
				TrimmedLine = TrimmedLine.Substring(i).TrimStart();

				// Storage for the tokenised line data.
				var TokenisedLine = new List<byte>(256);

				var CurrentState = TokeniserState.StartOfStatement;

				var LineToTokeniseQueue = new Queue<char>(TrimmedLine);

				bool InsideName = false;

				// Pull in tokens from the start of the string first.
				while (LineToTokeniseQueue.Count > 0) {

					// Handle some special cases.
					switch (LineToTokeniseQueue.Peek()) {
						case '&': // Scans the following hex number.
							InsideName = false;
							TokenisedLine.Add(BbcBasicEncoding.CharToByte(LineToTokeniseQueue.Dequeue()));
							while (LineToTokeniseQueue.Count > 0) {
								if (!LineToTokeniseQueue.Peek().IsValidInHexConstant()) break;
								TokenisedLine.Add(BbcBasicEncoding.CharToByte(LineToTokeniseQueue.Dequeue()));
							}
							break;
						case '"': // Scans following string constant.
							InsideName = false;
							TokenisedLine.Add(BbcBasicEncoding.CharToByte(LineToTokeniseQueue.Dequeue()));
							while (LineToTokeniseQueue.Count > 0) {
								char addedChar;
								TokenisedLine.Add(BbcBasicEncoding.CharToByte(addedChar = LineToTokeniseQueue.Dequeue()));
								if (addedChar == '"') break;
							}
							break;
						case ':': // Goes back to "start of statement" state.
							InsideName = false;
							TokenisedLine.Add(BbcBasicEncoding.CharToByte(LineToTokeniseQueue.Dequeue()));
							CurrentState = TokeniserState.StartOfStatement;
							break;
						case '*': // Prevents tokenising if at the start of a statement.
							InsideName = false;
							TokenisedLine.Add(BbcBasicEncoding.CharToByte(LineToTokeniseQueue.Dequeue()));
							if (CurrentState == TokeniserState.StartOfStatement) {
								while (LineToTokeniseQueue.Count > 0) {
									TokenisedLine.Add(BbcBasicEncoding.CharToByte(LineToTokeniseQueue.Dequeue()));
								}
							}
							break;
						default: // Normal tokeniser magic.

							var CurrentLine = new string(LineToTokeniseQueue.ToArray());

							var HandledToken = false;

							if (InsideName) {
								InsideName = false;
							} else {
								foreach (var PossibleKeyword in KeywordStringsLongestFirst) {
									if (CurrentLine.StartsWith(PossibleKeyword)) {
										// We have found a keyword match.
										var StringKeyword = StringToKeywordMap[PossibleKeyword];

										var CanInsertToken = true;

										// We can't insert the token if it is flagged as "conditional" and followed by alphanumeric text.
										if ((StringKeyword.Flags & KeywordFlags.Conditional) != KeywordFlags.None) {
											if (CurrentLine.Length > StringKeyword.Text.Length && char.IsLetterOrDigit(CurrentLine[StringKeyword.Text.Length])) {
												CanInsertToken = false;
											}
										}

										if (CanInsertToken) {

											// We're inserting the token. Remove the keyword from the queue.
											for (int j = 0; j < StringKeyword.Text.Length; ++j) LineToTokeniseQueue.Dequeue();

											// Grab the token value.
											var TokenValue = StringKeyword.Token;

											// Is is a pseudo-variable?
											if ((StringKeyword.Flags & KeywordFlags.PseudoVariable) != KeywordFlags.None) {
												if (CurrentState == TokeniserState.StartOfStatement) TokenValue += 0x40;
											}

											// Insert the token proper.
											TokenisedLine.Add(TokenValue);

											// If this is a FN/PROC, do not tokenise the name immediately following the FN/PROC token.
											if ((StringKeyword.Flags & KeywordFlags.FnProc) != KeywordFlags.None) {
												while (LineToTokeniseQueue.Count > 0 && LineToTokeniseQueue.Peek().IsValidInName()) {
													TokenisedLine.Add(BbcBasicEncoding.CharToByte(LineToTokeniseQueue.Dequeue()));
												}
											}

											// If this is a REMark, do not tokenise the rest of the line.
											if ((StringKeyword.Flags & KeywordFlags.Remark) != KeywordFlags.None) {
												while (LineToTokeniseQueue.Count > 0) {
													TokenisedLine.Add(BbcBasicEncoding.CharToByte(LineToTokeniseQueue.Dequeue()));
												}
											}

											// Handle line numbers.
											if ((StringKeyword.Flags & KeywordFlags.LineNumber) != KeywordFlags.None) {
												var LineNumberList = new Queue<char>();
												while (LineToTokeniseQueue.Count > 0 && LineToTokeniseQueue.Peek().IsValidInLineNumberList()) {
													LineNumberList.Enqueue(LineToTokeniseQueue.Dequeue());
												}
												ushort? ParsedLineNumber = null;
												while (LineNumberList.Count > 0) {
													var LineNumberChar = LineNumberList.Dequeue();
													if (LineNumberChar >= '0' && LineNumberChar <= '9') {
														ParsedLineNumber = (ushort)((ParsedLineNumber ?? 0) * 10);
														ParsedLineNumber += (ushort)(LineNumberChar - '0');
													} else {
														if (ParsedLineNumber.HasValue) {
															TokenisedLine.AddRange(ParsedLineNumber.Value.ToEncodedLineNumber());
															ParsedLineNumber = null;
														}
														TokenisedLine.Add(BbcBasicEncoding.CharToByte(LineNumberChar));
													}
												}
												if (ParsedLineNumber.HasValue) TokenisedLine.AddRange(ParsedLineNumber.Value.ToEncodedLineNumber());
											}

											// Update tokeniser state as appropriate.
											if ((StringKeyword.Flags & KeywordFlags.Start) != KeywordFlags.None) CurrentState = TokeniserState.StartOfStatement;
											if ((StringKeyword.Flags & KeywordFlags.Middle) != KeywordFlags.None) CurrentState = TokeniserState.MiddleOfStatement;
										} else {
											// We couldn't insert the token. Oh well. Just insert the token in full, I suppose.
											for (int j = 0; j < StringKeyword.Text.Length; ++j) {
												TokenisedLine.Add(BbcBasicEncoding.CharToByte(LineToTokeniseQueue.Dequeue()));
											}
										}
										HandledToken = true;
										break;
									}
								}
							}

							// We didn't manage to handle a token, oh well. Insert a character and try again.
							if (!HandledToken) {
								var ToEncode = LineToTokeniseQueue.Dequeue();
								CurrentState = TokeniserState.MiddleOfStatement;
								InsideName = (ToEncode < '0' || ToEncode > '9') && ToEncode.IsValidInName();
								TokenisedLine.Add(BbcBasicEncoding.CharToByte(ToEncode));								
							}
							break;							
					}
				}

				// Write the binary data to the list.
				Output.Add((byte)(TokenisedLine.Count + 4));
				Output.Add((byte)(LineNumber >> 0));
				Output.Add((byte)(LineNumber >> 8));
				Output.AddRange(TokenisedLine);
				Output.Add((byte)0x0D);

			}
			Output.Add(0x00); // 0-length line.
			Output.Add(0xFF); Output.Add(0xFF); // Line number 65535
			return Output.ToArray();
		}

		/// <summary>
		/// Renumber a program.
		/// </summary>
		/// <param name="program">The tokenised program.</param>
		/// <param name="start">The first line number to use.</param>
		/// <param name="step">The step between line numbers.</param>
		public static void Renumber(byte[] program, int start, int step) {
			if (step < 1) throw new ArgumentException("Step must be greater than zero.", "step");
			if (start < 1) throw new ArgumentException("Starting line number must be greater than zero.", "start");
			
			var RenumberedLines = new Dictionary<ushort, ushort>(); // Stores the old->new line number mapping.
			var FixableGotos = new List<KeyValuePair<int, ushort>>();

			ushort NewLineNumber = checked((ushort)start);
			for (int i = 0; i < program.Length; ) {
				int LineLength = program[i + 0];
				ushort LineNumber = program[i + 1];
				LineNumber += (ushort)(256 * program[i + 2]);
				if (LineNumber == 0xFFFF && LineLength == 0) break;
				
				if (!RenumberedLines.ContainsKey(LineNumber)) {
					RenumberedLines.Add(LineNumber, NewLineNumber);
				}

				program[i + 1] = (byte)(NewLineNumber & 0xFF);
				program[i + 2] = (byte)(NewLineNumber >> 8);

				try {
					NewLineNumber = checked((ushort)(NewLineNumber + step));
				} catch {
					throw new OverflowException("Line number went out of valid range.");
				}
				if (NewLineNumber == 0xFFFF) throw new OverflowException("Line number went out valid range.");

				for (int j = 3; j < LineLength - 1; ++j) {
					if (program[i + j] == 0x8D) FixableGotos.Add(new KeyValuePair<int, ushort>(i + j, LineNumber));
				}

				i += LineLength;
			}

			foreach (var Goto in FixableGotos) {
				var OldLine = program.DecodeLineNumber(Goto.Key);
				if (OldLine == 0xFFFF) continue;
				if (!RenumberedLines.ContainsKey(OldLine)) {
					throw new InvalidDataException(string.Format("Line {0} (referenced in line {1}) does not exist.", OldLine, Goto.Value));
				} else {
					var NewLine = RenumberedLines[OldLine].ToEncodedLineNumber();
					Array.Copy(NewLine, 0, program, Goto.Key, NewLine.Length);
				}
			}

		}

		public static string[] GetProceduresAndFunctions(byte[] program) {
			var Result = new List<string>(32);

			var E = new BbcBasicEncoding();

			for (int i = 0; i < program.Length; ) {
				int LineLength = program[i + 0];
				ushort LineNumber = program[i + 1];
				LineNumber += (ushort)(256 * program[i + 2]);
				if (LineNumber == 0xFFFF && LineLength == 0) break;

				for (int j = 3; j < LineLength; j++) {
					if (program[i + j] == 0xDD) {
						var ProcName = new StringBuilder(32);
						for (++j; j < LineLength && program[i + j] == ' '; ++j) ;
						string Prefix = null;
						if (j < (LineLength - 1) && (program[i + j] == 0xA4 || program[i + j] == 0xF2)) {
							Prefix = (program[i + j] == 0xA4 ? "FN" : "PROC");
							char c;
							for (++j; j < LineLength && (c = E.GetChars(program, i + j, 1)[0]).IsValidInName(); ++j) {
								ProcName.Append(c);
							}
						}
						if (ProcName.Length > 0) {
							Result.Add(Prefix + ProcName.ToString());
						}
					}
				}
		
				i += LineLength;
			}

			Result.Sort();

			return Result.ToArray();
		}

	}
}
